/*************************************************************************
**                                                                        
**   COPYRIGHT (c) 1999   Geotech Instruments, LLC
**
**                        10755 Sanden Drive
**                        Dallas, Texas 75238
**
**                        (214) 221-0000
**
**************************************************************************
**
** Module Name:   Timefix
** System:        A-900/A-900A
** Programmer:    Paul Akers
** Date Created:  11/02/99
**
**
** Module Description:
**
**    Fixes file time and date problems in A-900 or A-900A data files
**    caused by GPS WNRO problems.
**
**    Usage: timefix filename [/y] [/q]
**
**    Where: filename - File to fix, include drive and path if necessary.
**                      Note: Must be first parameter.
**           /y - Auto confirm modify file question with Yes.
**           /q - Quiet mode, does not display messages.
**
**    Returns:  0 if time correction and file modification where successful
**                (or time did not need to be corrected).
**             -1 on any type of error.
**
**    WARNING: This program MUST NOT be used on correctly dated data
**    files with year values of 1998 and earlier. If used on these files,
**    the time and date will be corrupted!
**
**    This program can be used on data files that should have a correct
**    year value of 1999 through 2015. It can be used on data files that
**    have bad WNRO time and date as well as on data files with correct
**    time and date in this range.
**
** Data Structures:
**
**
**
** Functions:
**
**
**
** Dependencies:
**
**
**
** History/Modifications:
**
**      Date   |   Programmer  |  Remarks/Change Description
**    MM/DD/YY |               |  
** -----------------------------------------------------------------------
**    
**
**
**
**
*************************************************************************/

#include <conio.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define  FALSE    0
#define  TRUE     1

/*
** Function prototypes
*/
short ConvertPostWNRODate (short *year,
                           short *month,
                           short *day,
                           short *hour,
                           short *min,
                           short *sec,
                           short *corrected);

short DayOfYear (short year, short month, short day);

/*
** GPS WNRO constants
*/
short cumdim[15] = {0, -59, -28, 0, 31, 61, 92, 122,
                    153, 184, 214, 245, 275, 306, 337};

short divday[5] = {0,0,20,21,6};

short posdaybias[5] = {0,0,11,9,10};

short negdaybias[5] = {0,0,20,21,0};

/*
** Day of month lookup table.
*/
short num_days_in_month[12] = {31,28,31,30,31,30,31,31,30,31,30,31};


/*************************************************************************
**
** MAIN
**
*************************************************************************/

short main(short argc, char *argv[])
   {
   FILE  *stream;
   short i;
   short ret = 0;
   fpos_t pos;
   short quiet_mode = FALSE;
   short confirm = FALSE;
   short help = FALSE;
   short modify = FALSE;
   char  rw_string[128];
   char  temp_string[4];
   short in_char;
   short year;
   short month;
   short day;
   short hour;
   short min;
   short sec;
   short doy;
   short corrected;
   char * char_ptr;
/*
** Check for command line options.
*/
   for (i = 0 ; i < argc ; i++)
      {
      if (_stricmp(argv[i], "/q") == 0)
         {
         quiet_mode = TRUE;
         }
      else if (_stricmp(argv[i], "/y") == 0)
         {
         confirm = TRUE;
         }
      else if (_stricmp(argv[i], "/?") == 0)
         {
         help = TRUE;
         }
      else
         {
         }
      }
/*
** Display sign on message.
*/
   if (!quiet_mode || help || (quiet_mode && !confirm))
      {
      printf("\n\rA-900 Data File GPS Time Correction\r\n");
      printf("Geotech Instruments, LLC\r\n");
      printf("Version 1.1\r\n\r\n");
      }

   if (help)
      {
      if (!quiet_mode)
         {
         printf("Useage: timefix filename [/y] [/q] [/?]\r\n\r\n");
         printf("        /y - Enable auto-confirm mode.\r\n");
         printf("        /q - Enable quiet mode.\r\n");
         printf("        /? - Display help.\r\n\r\n");
         printf("Program end.\r\n\r\n");
         }
      ret = 0;
      return(ret);
      }

   if (quiet_mode && !confirm)
      {
      printf("\r\nCannot run in QUIET (/q) mode without\r\n");
      printf("CONFIRM (/y) mode also enabled!.\r\n\r\n");
      printf("Program end.\r\n\r\n");
      ret = -1;
      return(ret);
      }
/*
** See if a filename was input.
*/
   if (argc > 1)
      {
/*
**    Try to open the file.
*/
      stream = fopen(argv[1], "r+b");

      if (stream != NULL)
         {
         if (!quiet_mode)
            {
            printf("File %s opened.\r\n\r\n", argv[1]);
            }
/*
**       Check to make sure this is an A-900 data file.
*/
         pos = 48;
         rw_string[0] = 0;
         if ((fsetpos(stream, &pos) == 0) &&
             (fgets(rw_string, 16, stream) != NULL))
            {
            if ((_stricmp(rw_string, "event9\r\n") == 0) ||
                (_stricmp(rw_string, "cal9  \r\n") == 0))
               {
/*
**             Read the time and date from the file header.
*/
               pos = 20;
               fsetpos(stream, &pos);
               fgets(rw_string, 32, stream);

               if (!quiet_mode)
                  {
                  printf("Current data file time is....: %s\r\n", rw_string);
                  }
/*
**             Convert time and date string to integer numbers.
*/
               strncpy(temp_string, &rw_string[0], 2);
               temp_string[2] = 0;
               hour = atoi(temp_string);

               strncpy(temp_string, &rw_string[3], 2);
               temp_string[2] = 0;
               min = atoi(temp_string);

               strncpy(temp_string, &rw_string[6], 2);
               temp_string[2] = 0;
               sec = atoi(temp_string);

               strncpy(temp_string, &rw_string[9], 2);
               temp_string[2] = 0;
               month = atoi(temp_string);

               strncpy(temp_string, &rw_string[12], 2);
               temp_string[2] = 0;
               day = atoi(temp_string);

               strncpy(temp_string, &rw_string[15], 2);
               temp_string[2] = 0;
               year = atoi(temp_string);
/*
**             Fix up the year for the conversion process.
*/
               if (year > 16)
                  {
                  year += 1900;
                  }
               else if (year < 16)
                  {
                  year += 2000;
                  }
               else  /* Year is 16  */
                  {
                  year += 2100;
                  }
/*
**             Convert the bad time and date to a good one.
*/
               if (ConvertPostWNRODate(&year, &month, &day,
                                       &hour, &min, &sec, &corrected))
                  {
/*
**                Calculate the day of year.
*/
                  doy = DayOfYear(year, month, day);
/*
**                Reduce year to two digits.
*/
                  if (year < 2000)
                     {
                     year -= 1900;
                     }
                  else
                     {
                     year -= 2000;
                     }
/*
**                Format a new time and date string.
*/
                  sprintf(rw_string, 
                          "%2.2d:%2.2d:%2.2d %2.2d/%2.2d/%2.2d \r\n",
                          hour, min, sec, month, day ,year);

                  if (!quiet_mode)
                     {
                     printf("Corrected data file time is..: %s",
                            rw_string);
                     printf("Corrected doy of year is.....: %3.3d\r\n\r\n",
                            doy);
                     }
/*
**                See if a time correction was made. We will only modify
**                the file if it was.
*/
                  if (corrected)
                     {
/*
**                   Ask to confirm file modification.
*/
                     if (confirm)
                        {
                        modify = TRUE;    
                        }
                     else
                        {
/*
**                      Flush keyboard.
*/
                        while (_kbhit())
                           {
                           _getch();
                           }
/*
**                      Loop for user input.
*/
                        while (1)
                           {
                           printf("Correct the data file (Y/N)? ");
                     
                           while (!_kbhit())
                              {
                              }

                           in_char = _getche();
                           printf("\r\n");

                           if ((in_char == 'n') || (in_char == 'N'))
                              {
    
                              if (!quiet_mode)
                                 {
                                 printf("\r\nData file not modifed.\r\n\r\n");
                                 }
                              break;
                              }

                           if ((in_char == 'y') || (in_char == 'Y'))
                              {
                              modify = TRUE;    
                              break;
                              }
                           }
                        }
/*
**                   Modify the file if given the OK.
*/                  
                     if (modify)
                        {
/* 
**                      Write the time and date string to the file header.
*/
                        pos = 20;
                        fsetpos(stream, &pos);
                        if (fputs(rw_string, stream) >= 0)
                           {
/* 
**                         Fix up the file name in the file header with
**                         the new day of year.
*/
                           pos = 0;
                           fsetpos(stream, &pos);
                           fgets(rw_string, 32, stream);
                           sprintf(&rw_string[8], "%3.3d \r\n", doy);
                           pos = 0;
                           fsetpos(stream, &pos);
                           if (fputs(rw_string, stream) >= 0)
                              {
/*
**                            Close the file.
*/
                              fclose(stream);
/*
**                            Rename the file with the new day of week extension.
*/
                              strcpy(rw_string, argv[1]);
                              char_ptr = strchr(rw_string, '.') + 1;
                              sprintf(char_ptr, "%3.3d", doy);
                              if (rename(argv[1], rw_string) == 0)
                                 {
                                 if (!quiet_mode)
                                    {
                                    printf("\r\nData file modified for correct time.\r\n\r\n");
                                    printf("Data file renamed to: %s\r\n\r\n", rw_string);
                                    }
                                 }
                              else
                                 {
                                 if (!quiet_mode)
                                    {
                                    printf("ERROR - Data file rename error!\r\n\r\n");
                                    }

                                 ret = -1;
                                 }
                              }
                           else
                              {
                              if (!quiet_mode)
                                 {
                                 printf("ERROR - Data file write error!\r\n\r\n");
                                 }

                              ret = -1;
                              fclose(stream);
                              }
                           }
                        else
                           {
                           if (!quiet_mode)
                              {
                              printf("ERROR - Data file write error!\r\n\r\n");
                              }

                           ret = -1;
                           fclose(stream);
                           }
                        }
                     }
                  else
                     {
                     if (!quiet_mode)
                        {
                        printf("Data file time and date is correct.\r\n");
                        printf("No modification necessary.\r\n\r\n");
                        }

                     fclose(stream);
                     }
                  }
               else
                  {
                  if (!quiet_mode)
                     {
                     printf("ERROR - Time/Date conversion error!\r\n\r\n");
                     }

                  ret = -1;
                  fclose(stream);
                  }
               }
            else
               {
               if (!quiet_mode)
                  {
                  printf("ERROR - Input file is not an A-900 data file!\r\n\r\n");
                  }

               ret = -1;
               fclose(stream);
               }
            }
         else
            {
            if (!quiet_mode)
               {
               printf("ERROR - Input file error!\r\n\r\n");
               }

            ret = -1;
            fclose(stream);
            }  
         }
      else
         {
         if (!quiet_mode)
            {
            printf("ERROR - Bad input file name!\r\n\r\n");
            }
         ret = -1;
         }
      }
   else
      {
      if (!quiet_mode)
         {
         printf("ERROR - No input file name!\r\n\r\n");
         }

      ret = -1;   
      }

   if (!quiet_mode)
      {
      printf("Program end.\r\n\r\n");
      }

   return(ret);
   }

/*************************************************************************
**
** Function Name: ConvertPostWNRODate
** Programmer:    Trimble
** Date Created:  October 14, 1996
**
**
** Function Description:
**
** Converts TAIP TM times and dates from message value to correct value.
** Works on all TAIP messages from all receivers through the end of year 2017.
** If values are already correct, they are passed back unchanged.
** Returns TRUE if date has been processed successfully.
** Returns FALSE if input date is not a possible TAIP TM date.
**
**
** Calling Convention:
**
**
**
** Input Parameter Name  |    Type      |   Description
** -----------------------------------------------------------------------
**
**
**
** Output Parameter Name |    Type      |   Description
** -----------------------------------------------------------------------
**
**
**
*************************************************************************/

short ConvertPostWNRODate (short *year,
                           short *month,
                           short *day,
                           short *hour,
                           short *min,
                           short *sec,
                           short *corrected)
   {
   short ctr, leapday;

   *corrected = FALSE;
/*
** Possible TM dates are all dates from 6 jan 1980 through the year 2017 and
** 55 days in the months of feb, mar, and apr of the year 2116
*/
   if ((*year) < 1999)
      {

      leapday = (*month == 2) && (*day == 29);

      /*
      ** compute apparent days since 1 mar 1976
      */
      (*day) = (*day) + 365*((*year) - 1976) + cumdim[(*month)];

      if ((*day) < 1407)
         {
         return FALSE;
         }

      /*
      ** then add add one (leap) day for every four years = 1461 days
      ** max four cycles; could be replaced by divide
      */
      ctr = (*day) - 1460;
      while (ctr > 0)
         {
         (*day)++;
         ctr -= 1460;
         }
      if (leapday)
         {
         (*day)--;
         }
      /*
      ** then add 1024 weeks for second GPS epoch and at the same time
      ** subtract 5*1461 to base time at 01 mar 1996 instead of 01 mar 1980
      ** -137 = (1024*7 - 5*1461)
      */
      (*day) -= 137;
      (*year) = 1996;

      /*
      ** (*day) is now days since 01 mar 1996
      ** (*month) is months since march
      ** (*year) = year
      ** max four cycles; could be replaced by divide
      */
      while ((*day) >= 1461)
         {
         (*day) -= 1461;
         (*year) += 4;
         }

      if((*day) == 0)
         {
         /*
         ** leap day - add another four years to (*year) & return
         */
         (*day) = 29;
         (*month) = 2;
         }
      else
         {
         /*
         ** max three cycles; could be replaced by divide
         */
         while ((*day) > 365)
            {
            (*day) -= 365;
            (*year) ++;
            }
         (*month) = 14;

         /*
         ** max twelve cycles; could be replaced by efficient sort
         */
         while ((*day) <= cumdim[(*month)])
            {
            (*month)--;
            }
         (*day) -= cumdim[(*month)];

         if ((*month) > 12)
            {
            (*year)++;
            (*month) -= 12;
            }

         }
      /* no change to hour, minute, second */
      *corrected = TRUE;
      return TRUE;
      }
   else if ((*year) == 2116)
   /*
   ** 8 weeks of strange TM messages between 22 aug and 15 oct 1999
   ** output dates will always be between 11 feb and 6 apr 2116
   ** both time and date are biased
   */
      {
#define TAIPSECBIAS 16
#define TAIPMINBIAS 28
#define TAIPHOURBIAS 6

      *year = 1999;
      if (*month > 4)
         {
         return FALSE;
         }
      /*
      ** compute correct hour, minute, second;
      ** no leap second will occur during this short time period
      */
      (*sec) -= TAIPSECBIAS;
      if ((*sec) < 0)
         {
         (*sec)+= 60;
         (*min)--;
         }
      (*min) -= TAIPMINBIAS;
      if ((*min) < 0)
         {
         (*min)+=60;
         (*hour)--;
         }
      (*hour) -= TAIPHOURBIAS;
      if ((*hour) < 0)
         {
         (*hour)+=24;
         (*day)--;
         }

      /* note - value of zero allowed for (*day) in the next equaitons */

      /* map into correct month & day */
      if ((*day) <= divday[(*month)])
         {
         (*day) += posdaybias[(*month)];
         (*month) += 6;
         }
      else
         {
         (*day) -= negdaybias[(*month)];
         (*month) += 7;
         }
      *corrected = TRUE;
      return TRUE;
      }
   else if ((*year) < 2016)
      {
      /*
      ** TAIP message OK
      */
      return TRUE;
      }
   return FALSE;
   }

/*************************************************************************
**
** Function Name: DayOfYear
** Programmer:    Paul Akers
** Date Created:  11/02/99
**
**
** Function Description:
**
**    Calculates Day of Year value.
**
** Calling Convention:
**
**
**
** Input Parameter Name  |    Type      |   Description
** -----------------------------------------------------------------------
**
**
**
** Output Parameter Name |    Type      |   Description
** -----------------------------------------------------------------------
**
**
**
*************************************************************************/

short DayOfYear (short year, short month, short day)
   {
   short i;
   short doy;
/*
** Figure Day of Year.
*/
   if (month == 1)
      {
      doy = day;
      }
   else
      {
      doy = 0;

      for (i = 0; i < month - 1; i++)
         {
         doy += num_days_in_month[i];
         }

      doy += day;

      if ((month > 2) && (year % 4 == 0)) doy++;
      }

   return doy;
   }
